/*
Copyright 2012 Nathanael Hübbe
Contact: nathanael.huebbe@informatik.uni-hamburg.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _GNU_SOURCE

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <pnetcdf.h>
#include <mpi.h>

#define handleError(call) do {\
	int error = call;\
	if(error) {\
		fprintf(stderr, "error code %d returned by " #call "\n", error);\
		return error;\
	}\
} while(0)

const int kDimensionCount = 3;
const int kVariableCount = 40;
const int kRecordCount = 20;
const MPI_Offset kDimensionSizes[] = {1, 20000, 100};
const long kDistributedDimension = 1;	//The index of the dimension, that is distributed among the processes (the source of parallelism).

// Use -r | --reduced to get reduced output (only the naked throughput in MiB/s is printed).
// Use -p | --partial to stop process 0 from writing anything, which will leave holes in the file
// Use -s | --synchronize to ensure that only one variable is written to at a time.
int main(int argc, char** argv){
	int outputFile;
	int ret, myProcId, procCount;
	int varIds[kVariableCount], dimIds[kDimensionCount];
	MPI_Offset startCoords[kDimensionCount], reducedSizes[kDimensionCount];
	long i, floatCount;
	char name[100];	//I know this is bad, but I can't use glibc on the blizzard, so there's no asprintf available.
	struct timeval startTime, endTime;
	double barrierTime = 0;
	bool reducedOutput = false, partialWrite = false, addBarrier = false;
	
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myProcId);
	MPI_Comm_size(MPI_COMM_WORLD, &procCount);
	
	for(i = 1; i < argc; i++) {
		if(!strcmp(argv[i], "-r") || !strcmp(argv[i], "--reduced")) reducedOutput = true;
		else if(!strcmp(argv[i], "-p") || !strcmp(argv[i], "--partial")) partialWrite = true;
		else if(!strcmp(argv[i], "-s") || !strcmp(argv[i], "--synchronize")) addBarrier = true;
		else {
			if(!myProcId) {
				printf("Unknown argument: %s\nSupported options:\n", argv[i]);
				printf("\t-p | --partial Process 0 does not write anything leaving holes in the file\n");
				printf("\t-r | --reduced Print only the naked throughput in MiB/s\n");
				printf("\t-s | --synchronize Ensure that only one variable is written to at a time by adding MPI_Barriers\n");
			}
			return -1;
		}
	}

//	if(!myProcId) printf("Using pnetcdf for output with %i processes.\n", procCount);
	handleError(gettimeofday(&startTime, NULL));
	ret = ncmpi_create(MPI_COMM_WORLD, "/work/ku0598/bench.nc", NC_WRITE|NC_64BIT_OFFSET, MPI_INFO_NULL, &outputFile);
//	ret = nc_set_fill(outputFile, NC_NOFILL, &intTrash); printf("Using NC_NOFILL\n");

	//write the dimension descriptions
	floatCount = 1;
	for(i = 0; i < kDimensionCount; i++) {
		if(snprintf(name, sizeof(name), "dimension_%li", i+1) <= 0) return -1;
		if(i) {
			handleError(ncmpi_def_dim(outputFile, name, kDimensionSizes[i], &dimIds[i]));
			floatCount *= kDimensionSizes[i];
		} else {
			handleError(ncmpi_def_dim(outputFile, name, NC_UNLIMITED, &dimIds[i]));
		}
	}

	//write the variable descriptions
	for(i = 0; i < kVariableCount; i++) {
		if(snprintf(name, sizeof(name), "variable_%li", i+1) <= 0) return -1;
		handleError(ncmpi_def_var(outputFile, name, NC_FLOAT, kDimensionCount, dimIds, &varIds[i]));
	}
	handleError(ncmpi_enddef(outputFile));
	handleError(ncmpi_begin_indep_data(outputFile));

	//fake some data
	MPI_Offset startCoord = myProcId*(long)kDimensionSizes[kDistributedDimension]/procCount;
	MPI_Offset stopCoord = (myProcId + 1)*(long)kDimensionSizes[kDistributedDimension]/procCount;
	long localFloatCount = (floatCount/kDimensionSizes[kDistributedDimension])*(stopCoord - startCoord);
	float* data = (float*)malloc(sizeof(float)*localFloatCount);
	for(i = 0; i < localFloatCount; i++) data[i] = (float)i;

	//write it many times
	if(partialWrite && !reducedOutput && !myProcId) printf("Process 0 is twiddling his thumbs...\n");
	for(i = 0; i < kDimensionCount; i++) {
		startCoords[i] = 0;
		reducedSizes[i] = kDimensionSizes[i];
	}
	startCoords[kDistributedDimension] = startCoord;
	reducedSizes[kDistributedDimension] = stopCoord - startCoord;
	for(i = 0; i < kRecordCount; i++) {
		startCoords[0] = i;
		long curVar;
		for(curVar = 0; curVar < kVariableCount; curVar++) {
			if(!partialWrite || myProcId) handleError(ncmpi_put_vara_float(outputFile, varIds[curVar], startCoords, reducedSizes, data));
			if(addBarrier) {
				struct timeval barrierStart, barrierEnd;
				if(!reducedOutput) handleError(gettimeofday(&barrierStart, NULL));
				MPI_Barrier(MPI_COMM_WORLD);
				if(!reducedOutput) {
					handleError(gettimeofday(&barrierEnd, NULL));
					barrierTime += (double)(barrierEnd.tv_sec - barrierStart.tv_sec) + (barrierEnd.tv_usec - barrierStart.tv_usec)/(double)1000000;
				}
			}
		}
	}

	handleError(ncmpi_close(outputFile));

	if(!myProcId) {
		handleError(gettimeofday(&endTime, NULL));
		double elapsedTime = (double)(endTime.tv_sec - startTime.tv_sec) + (endTime.tv_usec - startTime.tv_usec)/(double)1000000;
		double throughput = floatCount*kRecordCount*kVariableCount*4/(1024*1024*elapsedTime);
		if(reducedOutput) {
			printf("%.3f\n", throughput);
		} else {
			printf("Elapsed time: %.6fs\n", elapsedTime);
			printf("Speed: %.3fMiByte/s\n", throughput);
		}
	}
	if(addBarrier && !reducedOutput) {
		for(i = 0; i < procCount; i++) {
			MPI_Barrier(MPI_COMM_WORLD);	//Make sure, that this appears in the correct order.
			if(i == myProcId) printf("Process %d: %.6fs spent in MPI_Barrier()\n", myProcId, barrierTime);
		}
	}

	MPI_Finalize();
	
	return 0;
}
